// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Authors: Antonio Carzaniga <carzanig@cs.colorado.edu>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2001-2002 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: bitvector.cc,v 1.6 2002/09/29 06:47:50 carzanig Exp $
//
#include "fwdconf.h"
#include "ft_allocator.h"
#include "bitvector.h"

#include <iostream>

#ifndef HAVE_INLINE
#include "bitvector.icc"
#endif

//
// map of a position value for the bitvector
//
//       index 2     index 1     block   element
// ...+-----------+-----------+--------+---------+
//
bool ibitvector::test(size_t pos) const {
    if (pos > size) return false;
    size_t atom_mask = 1 << (pos & bv_atom_bits_mask);
    pos >>= bv_atom_bits_shift;
    size_t block_pos = pos & block_mask;
    pos >>= index_shift;
    const block * b = &first_block;
    
    if (pos) {
	const index * bi = first_block.up;
	if (bi == NULL) return false;

	size_t index_addr = pos & index_mask;
	unsigned char level = 0;

	while (pos >>= index_shift) {
	    bi = bi->up;
	    if (bi == NULL) return false;
	    ++level;
	    index_addr <<= index_shift;
	    index_addr |= (pos & index_mask);
	}
	while(level-- > 0) {
	    bi = bi->down[index_addr & index_mask].i;
	    if (bi == NULL) return false;
	    index_addr >>= index_shift;
	}
	b = bi->down[index_addr & index_mask].b;
	if (b == NULL) return false;
    }

    return b->elements[block_pos] & atom_mask;
}

bool ibitvector::set(size_t pos, FTAllocator & ftmemory) {
    if (pos > size) size = pos;
    size_t atom_mask = 1 << (pos & bv_atom_bits_mask);
    pos >>= bv_atom_bits_shift;
    size_t block_pos = pos & block_mask;
    pos >>= index_shift;

    block * b = &first_block;
    
    if (pos) {
	index * p_bi;
	index ** bi = &(first_block.up);
	size_t index_addr = 0;
	unsigned char level = 0;

	do {
	    ++level;
	    index_addr <<= index_shift;
	    index_addr |= (pos & index_mask);
	    pos >>= index_shift;
	    if (*bi == NULL) *bi = new (ftmemory) index(NULL);
	    p_bi = *bi;
	    bi = &(p_bi->up);
	} while (pos);
	
	index_or_block * iob = &(p_bi->down[index_addr & index_mask]);

	while (--level > 0) {
	    index_addr >>= index_shift;
	    if (iob->i == NULL) iob->i = new (ftmemory) index(p_bi);
	    p_bi = iob->i;
	    iob = &(p_bi->down[index_addr & index_mask]);
	}
	if (iob->b == NULL) iob->b = new (ftmemory)block(p_bi);
	b = iob->b;
    }
    if (b->elements[block_pos] & atom_mask) {
	return true;
    } else {
	b->elements[block_pos] |= atom_mask;
	++count;
	return false;
    }
}

//
// map of a position value for the bitvector
//
//       index 2     index 1     block   element
// ...+-----------+-----------+--------+---------+
//
void bitvector::set(const ibitvector & x) {
    count += set(elements, &(elements[element_size()]), 
		 x.first_block.elements, 
		 &(x.first_block.elements[ibitvector::block_size]));

    ibitvector::iterator i(x.first_block);
    const ibitvector::block * b;
    size_t end;
    while ((b = i.next_block()) != NULL) {
	if (i.element_address() >= element_size()) return;
	end = i.element_address() + ibitvector::block_size < element_size()
	    ? i.element_address() + ibitvector::block_size : element_size() ;
	count += set(&(elements[i.element_address()]), &(elements[end]),
		     b->elements, &(b->elements[ibitvector::block_size]));
    }
}
